# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

import wejax

project = "WeJAX"
version = wejax.__version__
release = wejax.__version__
author = wejax.__author__
copyright = wejax.__copyright__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.bibtex",
    "sphinx_copybutton",
    "myst_nb",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinc_mapping = {
    "python": ("https://docs.python.org/3", None),
    "jax": ("https://jax.readthedocs.io/en/latest", None),
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"

html_static_path = ["_static"]

html_theme_options = {
    "announcement": "WeJAX is under development!",
    "source_repository": "https://gitlab.com/lisamission/wejax",
    "source_branch": "main",
    "source_directory": "docs",
    "light_logo": "img/wejax-light.svg",
    "dark_logo": "img/wejax-dark.svg",
    "footer_icons": [
        {
            "name": "Gitlub",
            "url": "https://gitlab.com/lisamission/wejax",
            "html": """
                <svg xmlns="http://www.w3.org/2000/svg" stroke="currentColor"
                fill="currentColor" stroke-width="0" viewBox="0 0 512 512">
                    <path d="M503.5 204.6L502.8 202.8L433.1 21C431.7 17.5 429.2
                    14.4 425.9 12.4C423.5 10.8 420.8 9.9 417.9 9.6C415 9.3 412.2
                    9.7 409.5 10.7C406.8 11.7 404.4 13.3 402.4 15.5C400.5 17.6
                    399.1 20.1 398.3 22.9L351.3 166.9H160.8L113.7 22.9C112.9
                    20.1 111.5 17.6 109.6 15.5C107.6 13.4 105.2 11.7 102.5
                    10.7C99.9 9.7 97 9.3 94.1 9.6C91.3 9.9 88.5 10.8 86.1
                    12.4C82.8 14.4 80.3 17.5 78.9 21L9.3 202.8L8.5 204.6C-1.5
                    230.8-2.7 259.6 5 286.6C12.8 313.5 29.1 337.3 51.5
                    354.2L51.7 354.4L52.3 354.8L158.3 434.3L210.9 474L242.9
                    498.2C246.6 500.1 251.2 502.5 255.9 502.5C260.6 502.5 265.2
                    500.1 268.9 498.2L300.9 474L353.5 434.3L460.2 354.4L460.5
                    354.1C482.9 337.2 499.2 313.5 506.1 286.6C514.7 259.6 513.5
                    230.8 503.5 204.6z"/>
                </svg>
            """,
            "class": "",
        },
    ],
}


# -- Options for extensions --------------------------------------------------

typehints_defaults = "comma"
always_use_bars_union = True

autodoc_type_aliases = {
    "Iterable": "Iterable",
    "ArrayLike": "ArrayLike",
}

autosectionlabel_prefix_document = True

mathjax3_config = {
    "loader": {"load": ["[tex]/physics"]},
    "tex": {
        "packages": {"[+]": ["physics"]},
    },
}

bibtex_bibfiles = ["bibliography.bib"]
bibtex_default_style = "unsrt"

nb_execution_mode = "off"



"""Test the fim module."""

import jax
import jax.numpy as jnp
import numpy as np
from jax import Array
from jax._src.typing import ArrayLike

from wejax.fim import (
    inner_product,
    matched_filter_gaussian_fim,
    univariate_inner_product,
)


def test_univariate_inner_product() -> None:
    """Test the univariate_inner_product function."""

    a = jnp.array([1, 2, 3, 4])
    b = jnp.array([4, 3, 2, 1])
    df = 0.1
    noise_psd = jnp.array([1, 1, 1, 1])

    result = univariate_inner_product(a, b, noise_psd, df=df)
    expected = 4 * jnp.real(jnp.trapezoid(a * jnp.conj(b) / noise_psd, dx=df))

    assert result.shape == ()
    np.testing.assert_approx_equal(result, expected)

    # Test with a different df
    df = 0.2

    result = univariate_inner_product(a, b, noise_psd, df=df)
    expected = 4 * jnp.real(np.trapezoid(a * jnp.conj(b) / noise_psd, dx=df))

    assert result.shape == ()
    np.testing.assert_approx_equal(result, expected)

    # Test with a different noise_psd
    noise_psd = jnp.array([1, 2, 3, 4])

    result = univariate_inner_product(a, b, df=df, noise_psd=noise_psd)
    expected = 4 * jnp.real(jnp.trapezoid(a * jnp.conj(b) / noise_psd, dx=df))

    assert result.shape == ()
    np.testing.assert_approx_equal(result, expected)

    # Test with a different sum_func
    def jax_sum(a: ArrayLike, dx: float, axis: int = -1) -> Array:
        return jnp.sum(a, axis=axis) * dx

    result = univariate_inner_product(
        a, b, df=df, noise_psd=noise_psd, sum_func=jax_sum
    )
    expected = 4 * np.real(np.sum(a * np.conj(b) / noise_psd) * df)

    assert result.shape == ()
    np.testing.assert_approx_equal(result, expected)

    # Test with additional dimensions
    a = jnp.array([[1, 2], [3, 4]])
    b = jnp.array([[4, 3], [2, 1]])
    noise_psd = jnp.array([[1, 2], [3, 4]])

    result = univariate_inner_product(a, b, df=df, noise_psd=noise_psd)
    expected = 4 * jnp.real(jnp.trapezoid(a * jnp.conj(b) / noise_psd, dx=df, axis=-1))

    assert result.shape == (2,)
    np.testing.assert_array_almost_equal(result, expected)


def test_univariate_inner_product_shape() -> None:
    """Test the univariate_inner_product function with different shapes."""

    df = 0.1
    a = jnp.ones((10,))
    b = jnp.ones((10,))
    noise_cov = jnp.ones((10,))
    assert univariate_inner_product(a, b, noise_cov, df=df).shape == ()

    a = jnp.ones((10,))
    b = jnp.ones((10,))
    noise_cov = jnp.ones((1,))
    assert univariate_inner_product(a, b, noise_cov, df=df).shape == ()

    a = jnp.ones((50, 10))
    b = jnp.ones((50, 10))
    noise_cov = jnp.ones((10,))
    assert univariate_inner_product(a, b, noise_cov, df=df).shape == (50,)

    a = jnp.ones((20, 10))
    b = jnp.ones((10))
    noise_cov = jnp.ones((10,))
    assert univariate_inner_product(a, b, noise_cov, df=df).shape == (20,)


def test_inner_product() -> None:
    """Test the inner_product function."""

    a = jnp.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])
    b = jnp.array([[13, 14, 15, 16], [17, 18, 19, 20], [21, 22, 23, 24]])
    df = 0.1
    noise_psd = jnp.array([1, 1, 1, 1])
    noise_cov = jnp.tile(jnp.diag(noise_psd), (3, 1, 1))

    result = inner_product(a, b, noise_cov, df=df)
    expected = 4 * np.real(
        np.trapezoid(
            np.einsum("...fi,...fij,...fj->...f", a, noise_cov, np.conj(b)),
            dx=df,
            axis=-1,
        )
    )

    assert result.shape == ()
    np.testing.assert_approx_equal(result, expected)

    # Test with uncorrelated channels
    noise_psd = jnp.array([1, 2, 3, 4])
    noise_cov = jnp.tile(jnp.diag(noise_psd), (3, 1, 1))

    result = inner_product(a, b, noise_cov, df=df)
    inv_noise_psd = 1 / noise_psd
    inv_noise_cov = jnp.tile(jnp.diag(inv_noise_psd), (3, 1, 1))
    expected = 4 * np.real(
        np.trapezoid(
            np.einsum("...fi,...fij,...fj->...f", a, inv_noise_cov, np.conj(b)),
            dx=df,
            axis=-1,
        )
    )

    assert result.shape == ()
    np.testing.assert_approx_equal(result, expected)

    # Test with full covariance matrix
    np_noise_cov = np.zeros((3, 4, 4), dtype=complex)
    np_noise_cov[:] += np.diag([1, 2, 3, 4])
    np_noise_cov[:, 0, 1] = 0.5 + 1j
    np_noise_cov[:, 1, 0] = 0.5 - 1j
    np_noise_cov[:, 1, 2] = 0.1 + 0.3j
    np_noise_cov[:, 2, 1] = 0.1 - 0.3j
    noise_cov = jnp.asarray(np_noise_cov)

    result = inner_product(a, b, noise_cov, df=df)
    inv_noise_cov = jnp.linalg.inv(noise_cov)
    expected = 4 * np.real(
        np.trapezoid(
            np.einsum("...fi,...fij,...fj->...f", a, inv_noise_cov, np.conj(b)),
            dx=df,
            axis=-1,
        )
    )

    assert result.shape == ()
    np.testing.assert_approx_equal(result, expected)

    # Test with the inverse covariance matrix given

    result = inner_product(a, b, inv_noise_cov, df=df, should_invert_noise_cov=False)

    assert result.shape == ()
    np.testing.assert_approx_equal(result, expected)

    # Test with a different sum_func
    def jax_sum(a: Array, dx: float, axis: int = -1) -> Array:
        return jnp.sum(a, axis=axis) * dx

    result = inner_product(a, b, df=df, noise_cov=noise_cov, sum_func=jax_sum)
    expected = 4 * np.real(
        np.sum(
            np.einsum("...fi,...fij,...fj->...f", a, inv_noise_cov, np.conj(b)),
            axis=-1,
        )
        * df
    )

    assert result.shape == ()
    np.testing.assert_approx_equal(result, expected)

    # Test with additional dimensions
    a = jnp.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    b = jnp.array([[[8, 7], [6, 5]], [[4, 3], [2, 1]]])
    noise_psd = jnp.array([1, 2])
    noise_cov = jnp.tile(jnp.diag(noise_psd), (2, 2, 1, 1))

    result = inner_product(a, b, noise_cov, df=df)
    inv_noise_cov = jnp.linalg.inv(noise_cov)
    expected = 4 * np.real(
        np.trapezoid(
            np.einsum("...fi,...fij,...fj->...f", a, inv_noise_cov, np.conj(b)),
            dx=df,
            axis=-1,
        )
    )

    assert result.shape == (2,)
    np.testing.assert_array_almost_equal(result, expected)


def test_inner_product_shape() -> None:
    """Test the inner_product function with different shapes."""

    df = 0.1
    a = jnp.ones((10, 10))
    b = jnp.ones((10, 10))
    noise_cov = jnp.ones((10, 10))
    assert inner_product(a, b, noise_cov, df=df).shape == ()

    a = jnp.ones((10, 10))
    b = jnp.ones((10, 10))
    noise_cov = jnp.ones((1, 10, 10))
    assert inner_product(a, b, noise_cov, df=df).shape == ()

    a = jnp.ones((50, 10, 10))
    b = jnp.ones((50, 10, 10))
    noise_cov = jnp.ones((10, 10))
    assert inner_product(a, b, noise_cov, df=df).shape == (50,)

    a = jnp.ones((20, 10, 10))
    b = jnp.ones((10, 10))
    noise_cov = jnp.ones((10, 10))
    assert inner_product(a, b, noise_cov, df=df).shape == (20,)


def test_matched_filter_gaussian_fim_shape() -> None:
    """Test the matched_filter_gaussian_fim function with different shapes."""

    # Mock a Jacobian function
    n_freq = 10
    n_channels = 3
    n_params = 2

    def jac(a: float, b: float) -> Array:
        """Mock Jacobian function."""
        jac_shape = (n_freq, n_channels, n_params)
        return (a + b) * jnp.ones(jac_shape)

    df = 0.1
    a = 1.0
    b = 1.0
    noise_cov = jnp.ones((n_channels, n_channels))
    fim = matched_filter_gaussian_fim(jac, noise_cov, df=df)
    assert fim(a, b).shape == (n_params, n_params)

    a = jnp.asarray(1.0)
    b = jnp.asarray(1.0)
    noise_cov = jnp.ones((n_freq, n_channels, n_channels))
    fim = matched_filter_gaussian_fim(jac, noise_cov, df=df)
    assert fim(a, b).shape == (n_params, n_params)

    a = jnp.ones((50,))
    b = jnp.ones((50,))
    noise_cov = jnp.ones((n_freq, n_channels, n_channels))
    vjac = jax.vmap(jac, in_axes=(0, 0))
    fim = matched_filter_gaussian_fim(vjac, noise_cov, df=df)
    assert fim(a, b).shape == (50, n_params, n_params)

    a = jnp.ones(
        (
            20,
            50,
        )
    )
    b = jnp.ones((50,))
    a, b = jnp.broadcast_arrays(a, b)
    noise_cov = jnp.ones((n_freq, n_channels, n_channels))
    vvjac = jax.vmap(vjac, in_axes=(0, 0))
    fim = matched_filter_gaussian_fim(vvjac, noise_cov, df=df)
    assert fim(a, b).shape == (20, 50, n_params, n_params)



r"""
Fisher information matrix
=========================

Use this module to compute Fisher Information Matrices (FIM) for set of
observables, assuming a Gaussian-noise matched filter-like likelihood.

Noise-weighted inner products
-----------------------------

Noise-weighted inner products are used to compute the Fisher Information Matrix.
They are defined as

.. math::

    (a|b) = 4 \Re \int_{f_{\min}}^{f_{\max}} \tilde{a}(f) \Sigma^{-1}(f)
    \tilde{b}^+(f) \, \mathrm{d}f,

where :math:`\tilde{a}(f)` and :math:`\tilde{b}(f)` are single or multi-channel
complex-valued Fourier transforms of the observables :math:`a(t)` and
:math:`b(t)`. The noise covariance matrix :math:`\Sigma(f)` characterizes the
noise correlations between the various observable channels.

.. autofunction:: univariate_inner_product

.. autofunction:: inner_product

FIM computation
---------------

Use the :func:`matched_filter_gaussian_fim` function to compute the Fisher
Information Matrix from a Jacobian function. The function returns another
function, which can evaluate the Fisher Information Matrix for a set of
parameters.

.. autofunction:: matched_filter_gaussian_fim

.. autoclass:: FIMFunction
    :members:

"""

from __future__ import annotations

from typing import Callable, Protocol

import jax.numpy as jnp
from jax import Array
from jax._src.typing import ArrayLike as JaxArrayLike

from wejax.jacobian import Jacobian


def univariate_inner_product(
    a: Array,
    b: Array,
    noise_psd: Array,
    *,
    df: float,
    sum_func: Callable = jnp.trapezoid,
) -> Array:
    r"""Compute the noise-weighted inner product of two univariate observables.

    The inner product is computed in the frequency domain, and for two
    univariate real-valued observables. The inner product is defined as

    .. math::

        (a|b) = 4 \Re \int_0^\infty \frac{\tilde{a}(f) \tilde{b}^*(f)}{S(f)} \,
        \mathrm{d}f,

    where :math:`\tilde{a}(f)` and :math:`\tilde{b}(f)` are the Fourier
    transforms of the observables :math:`a(t)` and :math:`b(t)`, respectively,
    and :math:`S(f)` is the one-sided power spectral density of the noise.

    Note that we assume here that the noise is stationary, such that there is no
    correlation between the noise at different frequencies; ie., the noise power
    spectral density characterizes the noise power at each frequency.

    For multivariate observables, use :func:`inner_product` instead.

    Example
    -------
    >>> result = univariate_inner_product(a, b, my_psd, df=df)

    Parameters
    ----------
    a, b : Array of shape ``(..., Nf)``
        Discrete Fourier transforms for the observables to compute the
        inner product of. Since the signals are assumed to be real, we only need
        the positive frequencies, as returned by :func:`jax.numpy.fft.rfft`.

        Here, ``Nf`` is the number of frequencies and ``...`` are any additional
        dimensions.
    noise_psd : Array of shape ``(..., Nf)``
        Noise covariance matrix [/Hz].

        Here ``Nf`` is the number of frequencies and ``...`` are any additional
        dimensions.
    df
        Frequency spacing [Hz].
    sum_func : Callable, default :func:`jax.numpy.trapezoid`
        Function to use to sum over the frequencies.

        The function should take as first positional argument an JAX Array, and
        should accept the keyword arguments `dx` and `axis`. It should return an
        array of shape ``(...)``, where ``...`` are any additional dimensions.

        We suggest using :func:`jax.numpy.trapezoid`, or :func:`jax.numpy.sum`.

    Returns
    -------
    Array of shape ``(...)``
        Noise-weighted inner product.
    """
    # Convert to JAX arrays
    a = jnp.asarray(a)  # (..., Nf)
    b = jnp.asarray(b)  # (..., Nf)
    a, b = jnp.broadcast_arrays(a, b)  # (..., Nf)
    noise_psd = jnp.asarray(noise_psd)  # (..., Nf)

    # Compute the integrand
    integrand = a * jnp.conj(b) / noise_psd  # (..., Nf)

    # Sum over the frequencies
    return 4 * jnp.real(sum_func(integrand, dx=df, axis=-1))  # (...)


def inner_product(
    a: Array,
    b: Array,
    noise_cov: Array,
    *,
    df: float,
    should_invert_noise_cov: bool = True,
    sum_func: Callable = jnp.trapezoid,
) -> Array:
    r"""Compute the noise-weighted inner product of two sets of observables.

    The inner product is computed in the frequency domain, and for two
    sets of real-valued observables. The inner product is defined as

    .. math::

        (a|b) = 4 \Re \int_0^\infty \tilde{a}(f) \Sigma^{-1}(f)
        \tilde{b}^+(f) \, \mathrm{d}f,

    where :math:`\tilde{a}(f)` and :math:`\tilde{b}(f)` are the Fourier
    transforms of the sets of observables :math:`a(t)` and :math:`b(t)`,
    respectively, and :math:`\Sigma(f)` is the noise covariance matrix.

    By default, the inverse of the noise covariance matrix is computed and used
    to compute the inner product. If you want to provide the inverse noise
    covariance matrix directly, set `should_invert_noise_cov` to `False`.

    Note that we assume here that the noise is stationary, such that there is no
    correlation between the noise at different frequencies (i.e., the noise
    covariance matrix is frequency-dependent and only characterize the noise
    correlations between the various observables channels).

    For univariate observables, use :func:`univariate_inner_product` instead.

    Example
    -------
    >>> result = inner_product(a, b, noise_cov=noise_cov, df=df)

    Parameters
    ----------
    a, b : Array of shape ``(..., Nf, Nc)``
        Discrete Fourier transforms for the observables to compute the
        inner product of. Since the signals are assumed to be real, we only need
        the positive frequencies, as returned by :func:`jax.numpy.fft.rfft`.

        Here, ``Nf`` is the number of frequencies, ``Nc`` is the number of
        observable channels, and ``...`` are any additional dimensions.
    noise_cov : Array of shape ``(..., Nf, Nc, Nc)``
        Noise covariance matrix [/Hz].

        Here ``Nf`` is the number of frequencies, ``Nc`` is the number of
        observable channels, and ``...`` are any additional dimensions.
    df
        Frequency spacing [Hz].
    should_invert_noise_cov : bool, optional
        If `True`, the inverse of the noise covariance matrix is computed and
        used to compute the inner product. If `False`, the provided
        `noise_cov` is used directly as the inverse noise covariance matrix.
    sum_func : Callable, default :func:`jax.numpy.trapezoid`
        Function to use to sum over the frequencies.

        The function should take as first positional argument an JAX Array, and
        should accept the keyword arguments `dx` and `axis`. It should return an
        array of shape ``(...)``, where ``...`` are any additional dimensions.

        We suggest using :func:`jax.numpy.trapezoid`, or :func:`jax.numpy.sum`.

    Returns
    -------
    Array of shape ``(...)``
        Noise-weighted inner product.
    """
    # Inverse noise covariance matrix if needed
    if should_invert_noise_cov:
        inv_noise_cov = jnp.linalg.inv(noise_cov)  # (..., Nf, Nc, Nc)
    else:
        inv_noise_cov = noise_cov  # (..., Nf, Nc, Nc)

    # Convert to JAX arrays
    a = jnp.asarray(a)  # (..., Nf, Nc)
    b = jnp.asarray(b)  # (..., Nf, Nc)
    a, b = jnp.broadcast_arrays(a, b)  # (..., Nf, Nc)

    # Contract over the channel axis
    subscripts = "...i, ...ij, ...j -> ..."
    contracted = jnp.einsum(subscripts, a, inv_noise_cov, jnp.conj(b))

    # Sum over the frequencies
    summed = 4 * jnp.real(sum_func(contracted, dx=df, axis=-1))  # (...)
    return summed


class FIMFunction(Protocol):  # pylint: disable=too-few-public-methods
    """Protocol for Fisher Information Matrices functions.

    Parameters
    ----------
    *args : Array-like
        Model parameters as positional arguments.

    Returns
    -------
    Array of shape ``(..., Np, Np)``
        The Fisher Information Matrix evaluated at ``args``.

        Here ``Np`` is the number of parameters, and ``...`` are any additional
        dimensions returned by the Jacobian function.
    """

    def __call__(self, *args: JaxArrayLike) -> Array: ...


def matched_filter_gaussian_fim(
    jacobian: Jacobian,
    noise_cov: Array,
    *,
    df: float,
    should_invert_noise_cov: bool = True,
    sum_func: Callable = jnp.trapezoid,
) -> FIMFunction:
    r"""Return a function that computes the Fisher Information Matrix.

    Compute the Fisher Information Matrix for a template function, assuming a
    stationary Gaussian noise. The Fisher Information Matrix is given by

    .. math::

        I_{ij}(\theta) = \left( \frac{\partial h}{\partial \theta_i} \middle|
        \frac{\partial h}{\partial \theta_j} \right).

    The returned function can be called with a set of parameters, and will
    return the Fisher Information Matrix evaluated at those parameters.

    Example
    -------
    >>> fim = matched_filter_gaussian_fim(jacobian, noise_cov=noise_cov, df=df)
    >>> result_1 = fim(1.0, 2.0)
    >>> result_2 = fim(3.0, 4.0)

    The returned FIM function can be vectorized for efficient computation of
    multiple sets of parameters using :func:`jax.vmap` (see `JAX documentation
    <https://docs.jax.dev/en/latest/_autosummary/jax.vmap.html>`_).

    >>> vmap_fim = jax.vmap(fim, in_axes=(0, 0), out_axes=0)

    Parameters
    ----------
    jacobian : :class:`Jacobian`
        Jacobian function.
    noise_cov : Array of shape ``(..., Nf, Nc, Nc)``
        Noise covariance matrix [/Hz].

        Here ``Nf`` is the number of frequencies, ``Nc`` is the number of
        observable channels, and ``...`` are any additional dimensions.
    df
        Frequency spacing [Hz].
    should_invert_noise_cov : bool, optional
        If `True`, the inverse of the noise covariance matrix is computed and
        used to compute the inner product. If `False`, the provided
        `noise_cov` is used directly as the inverse noise covariance matrix.
    sum_func : Callable, default :func:`jax.numpy.trapezoid`
        Function to use to sum over the frequencies.

        The function should take as first positional argument an JAX Array, and
        should accept the keyword arguments `dx` and `axis`. It should return an
        array of shape ``(...)``, where ``...`` are any additional dimensions.

        We suggest using :func:`jax.numpy.trapezoid`, or :func:`jax.numpy.sum`.

    Returns
    -------
    :class:`FIMFunction`
        A function that computes the Fisher Information Matrix for a set of
        parameters.
    """

    def fim(*args: JaxArrayLike) -> Array:

        # Evaluate the Jacobian
        jacobian_val = jacobian(*args)  # (..., Nf, Nc, Np)
        jacobian_val = jnp.moveaxis(jacobian_val, -1, -3)  # (..., Np, Nf, Nc)

        # Compute inner product column by column
        fim_columns = [
            inner_product(
                jacobian_val,
                jacobian_val[..., [i], :, :],
                noise_cov,
                df=df,
                sum_func=sum_func,
                should_invert_noise_cov=should_invert_noise_cov,
            )  # (..., Np)
            for i in range(len(args))
        ]

        return jnp.stack(fim_columns, axis=-1)  # (..., Np, Np)

    return fim



r"""
Estimation errors
=================

This module provides functions to compute and plot error covariance and
correlation matrices.

Error covariance matrix
-----------------------

The errors on the estimated parameters can be characterized by their covariance
matrix :math:`\Sigma`, which is defined as the inverse of the Fisher information
matrix,

.. math::

    \mathrm{Cov}(\theta_i, \theta_j) = \Sigma_{ij} = \left( I^{-1} \right)_{ij}.

The diagonal elements of :math:`\Sigma` provide the variances of the estimated
parameters, and the off-diagonal elements describe the correlations.

.. autofunction:: covariance_matrix

.. autofunction:: stddevs

Correlation matrix
------------------

The correlation matrix :math:`P` is defined as the normalized covariance matrix,
i.e., the matrix of the standard deviations and the correlation coefficients.

.. autofunction:: correlation_matrix

Plotting
--------

You can plot the error covariance matrix using the function
:func:`wejax.errors.plot_covariance_matrix`.

.. image:: _static/img/correlation-matrix.png
    :alt: Example correlation matrix
    :align: center

.. autofunction:: plot_correlation_matrix

"""

from itertools import product
from typing import Sequence

import jax.numpy as jnp
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from jax import Array
from jax._src.typing import ArrayLike
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


def covariance_matrix(
    fim: ArrayLike,
    *,
    argnums: Sequence[int] | None = None,
    rtol: float | None = None,
) -> Array:
    """Compute the error covariance matrix from the Fisher information matrix.

    If the matrix is singular, use ``argnums`` to select the parameters for
    which the error covariance matrix is computed. If ``rtol`` is not ``None``,
    the Moore-Penrose pseudo-inverse is computed.

    Examples
    --------
    >>> cov = covariance_matrix(fim)

    Fixing some parameters yields a different covariance matrix:

    >>> cov2 = covariance_matrix(fim, argnums=(0, 1))
    >>> np.testing.assert_array_almost_equal(cov2, cov[..., :2, :2])
    False

    Parameters
    ----------
    F : Array-like of shape ``(..., Np, Np)``
        Fisher information matrix. Here, ``Np`` is the number of parameters.
    argnums : Sequence of int or None, optional
        Indices of the parameters for which the error covariance matrix is
        computed. By default, the error covariance matrix is computed for all
        parameters.
    rtol : float or None, optional
        Cutoff parameter for small singular values.

        Singular values smaller than ``rtol`` times the largest singular value
        are considered zero and the function returns the Moore-Penrose
        pseudo-inverse. If ``None``, the exact inverse is computed.

    Returns
    -------
    Array of shape ``(..., Na, Na)``
        Error covariance matrix. Here, ``Na`` is the number of parameters for
        which the error covariance matrix is computed (length of ``argnums``).
    """
    fim = jnp.asarray(fim)

    # By default, compute error covariance matrix for all parameters
    if argnums is None:
        argnums = list(range(fim.shape[-1]))

    # Check indices are valid
    if not all(0 <= i < fim.shape[-1] for i in argnums):
        raise ValueError("Invalid indices in argnums.")

    # Compute the error covariance matrix
    restricted_fim = fim[..., argnums, :][..., argnums]
    if rtol is None:
        error_covariance = jnp.linalg.inv(restricted_fim)
    else:
        error_covariance = jnp.linalg.pinv(restricted_fim, rtol=rtol, hermitian=True)

    return error_covariance


def _invert_fim_if_needed(
    *, fim: ArrayLike | None = None, cov: ArrayLike | None = None
) -> Array:
    """Invert FIM to get covariance matrix if not provided.

    Parameters
    ----------
    fim : Array-like of shape ``(..., Np, Np)`` or None
        Fisher information matrix. Here, ``Np`` is the number of parameters.
    cov : Array-like of shape ``(..., Np, Np)`` or None
        Error covariance matrix. Here, ``Np`` is the number of parameters.

    Returns
    -------
    Array of shape ``(..., Np, Np)``
        Error covariance matrix.

    Raises
    ------
    ValueError
        If neither ``fim`` nor ``cov`` is provided.
    ValueError
        If both ``fim`` and ``cov`` are provided.
    """
    if cov is None and fim is None:
        raise ValueError("Either cov or fim must be provided.")
    if cov is not None and fim is not None:
        raise ValueError("Only one of cov and fim must be provided.")

    # Invert FIM to get covariance matrix if not provided
    if fim is not None:
        cov = covariance_matrix(fim)
    assert cov is not None

    return jnp.asarray(cov)


def stddevs(*, fim: ArrayLike | None = None, cov: ArrayLike | None = None) -> Array:
    """Compute parameter standard deviations.

    You can provide the noise covariance matrix as input, to avoid inverting
    the Fisher information matrix.

    Examples
    --------
    >>> a = stddevs(fim=fim)
    >>> cov = covariance_matrix(fim)
    >>> b = stddevs(cov=cov)
    >>> np.testing.assert_array_almost_equal(a, b)
    True

    Parameters
    ----------
    fim : Array-like of shape ``(..., Np, Np)`` or None
        Fisher information matrix. Here, ``Np`` is the number of parameters.
    cov : Array-like of shape ``(..., Np, Np)`` or None
        Error covariance matrix. Here, ``Np`` is the number of parameters.

    Returns
    -------
    Array of shape ``(..., Np)``
        Standard deviations.
    """
    cov = _invert_fim_if_needed(fim=fim, cov=cov)
    return jnp.sqrt(jnp.diagonal(cov, axis1=-2, axis2=-1))


def correlation_matrix(
    *, fim: ArrayLike | None = None, cov: ArrayLike | None = None
) -> Array:
    """Compute the correlation matrix (normalized covariance matrix).

    You can provide the noise covariance matrix as input, to avoid inverting
    the Fisher information matrix.

    Examples
    --------
    >>> a = correlation_matrix(fim=fim)
    >>> cov = covariance_matrix(fim)
    >>> b = correlation_matrix(cov=cov)
    >>> np.testing.assert_array_almost_equal(a, b)
    True

    Parameters
    ----------
    fim : Array-like of shape ``(..., Np, Np)`` or None
        Fisher information matrix. Here, ``Np`` is the number of parameters.
    cov : Array-like of shape ``(..., Np, Np)`` or None
        Error covariance matrix. Here, ``Np`` is the number of parameters.

    Returns
    -------
    Array of shape ``(..., Np, Np)``
        Correlation matrix.
    """
    cov = _invert_fim_if_needed(fim=fim, cov=cov)

    # Normalize covariance matrix by the product of standard deviations
    # We use the outer product to compute the denominator efficiently
    std_devs = stddevs(cov=cov)
    outer_std_devs = jnp.einsum("...i, ...j -> ...ij", std_devs, std_devs)
    corr = cov / outer_std_devs

    # Replace diagonal elements with standard deviations
    diag = jnp.diag(std_devs)
    diag = jnp.broadcast_to(diag, corr.shape)
    corr = jnp.where(diag, diag, corr)

    return corr


def plot_correlation_matrix(
    corr: ArrayLike,
    *,
    ax: plt.Axes | None = None,
    param_names: Sequence[str] | None = None,
    param_units: Sequence[str] | None = None,
    title: str = "Correlation matrix",
    cmap: str = "bwr",
    show_cbar: bool = True,
    font_size: float = 9,
    label_rotation: float = 0,
) -> tuple[plt.Figure, plt.Axes]:
    r"""Plot the correlation matrix.

    Examples
    --------
    >>> corr = correlation_matrix(fim=fim)
    >>> names = ["$f_0$", "$d_L$", "$\phi_0$"]
    >>> units = ["Hz", "Mpc", "rad"]
    >>> fig, ax = plot_correlation_matrix(corr, param_names=names, param_units=units)
    >>> fig.savefig("correlation_matrix.pdf", bbox_inches="tight")

    Parameters
    ----------
    corr : Array-like of shape ``(Np, Np)``
        Correlation matrix. Here, ``Np`` is the number of parameters.
    ax : Axes or None, optional
        Matplotlib axes. If None, a new figure is created.
    param_names : Sequence of ``Np`` str or None, optional
        Parameter names. By default, the parameters are labeled with their
        indices.
    param_units : Sequence of ``Np`` str or None, optional
        Parameter units. By default, the parameters are labeled without units.
    title : str, optional
        Figure title.
    cmap : str, optional
        Colormap. See `matplotlib colormaps
        <https://matplotlib.org/stable/gallery/color/colormap_reference.html>`_.
    show_cbar: bool, optional
        If True, show colorbar.
    font_size : float, optional
        Font size.
    label_rotation : float, optional
        Rotation angle for x-axis parameter labels.

    Returns
    -------
    Figure
        Matplotlib figure.
    Axes
        Matplotlib axes.
    """
    # pylint: disable=consider-using-f-string

    # Create figure and axes if not provided
    if ax is None:
        fig, ax = plt.subplots()
    else:
        assert isinstance(ax.figure, plt.Figure)
        fig = ax.figure

    # Check correlation matrix shape
    corr = jnp.asarray(corr)
    n_params = corr.shape[0]

    # Configure axes and add labels
    _setup_axes(
        ax,
        n_params=n_params,
        param_names=param_names,
        param_units=param_units,
        title=title,
        font_size=font_size,
        label_rotation=label_rotation,
    )

    # Plot color squares and values
    cmap_obj = plt.cm.get_cmap(cmap)
    for i, j in product(range(n_params), repeat=2):

        # Skip upper triangle
        if i > j:
            continue

        value = corr[i, j]
        latex = (
            _format_number_as_latex(value)
            if i == j
            else r"${\mathrm{" + "{:0.2f}".format(float(value)) + r"}}$"
        )

        center = (0.5 + i, 0.5 + j)
        rect_anchor = (center[0] - 0.5, center[1] - 0.5)
        rect = plt.Rectangle(rect_anchor, 1, 1, fill=True, linewidth=0)
        text = plt.Text(*center, latex, ha="center", va="center", size=font_size)

        if i == j:
            rect.set_color("lightgrey")
            text.set_color("black")
        if i < j:
            text_color = "black" if abs(value) < 0.5 else "white"
            text.set_color(text_color)
            text.set_fontweight("bold")
            rect_color = cmap_obj(0.5 + 0.5 * value)
            assert isinstance(rect_color, tuple)
            rect.set_color(rect_color)

        ax.add_patch(rect)
        ax.add_artist(text)

    ax.set_aspect("equal")

    if show_cbar:
        norm = mpl.colors.Normalize(vmin=-1, vmax=1)
        scalar_map = mpl.cm.ScalarMappable(norm=norm, cmap=cmap)
        cax = inset_axes(ax, width="5%", height="30%", loc="upper right", borderpad=1.5)
        cbar = plt.colorbar(scalar_map, cax=cax, orientation="vertical")
        cbar.ax.tick_params(labelsize=font_size)
        cbar.ax.yaxis.set_ticks_position("left")
        cbar.set_label("Correlation", fontsize=font_size, labelpad=5)

    return fig, ax


def _setup_axes(
    ax: plt.Axes,
    *,
    param_names: Sequence[str] | None,
    param_units: Sequence[str] | None,
    title: str,
    font_size: float,
    label_rotation: float,
    n_params: int,
) -> None:
    """Setup axes and add labels for the correlation matrix plot.

    Parameters
    ----------
    ax : Axes
        Matplotlib axes.
    param_names : Sequence of ``n_params`` str or None
        Parameter names.
    param_units : Sequence of ``n_params`` str or None
        Parameter units.
    title : str
        Figure title.
    font_size : float
        Font size.
    label_rotation : float
        Rotation angle for x-axis parameter labels.
    n_params : int
        Number of parameters.
    """
    # Check if parameter names and units are provided
    if param_names is None:
        param_names = [f"$p_{i}$" for i in range(n_params)]
    if len(param_names) != n_params:
        raise ValueError("Number of parameter names must match the matrix size.")
    if param_units is None:
        param_units = [""] * n_params
    if len(param_units) != n_params:
        raise ValueError("Number of parameter units must match the matrix size.")

    # Configure axes and labels
    ax.set_xlim(0, n_params)
    ax.set_ylim(0, n_params)
    ax.invert_yaxis()  # so it goes from top to bottom
    ax.set_xticks(np.arange(n_params) + 0.5)  # ticks in the middle of the squares
    ax.set_yticks(np.arange(n_params) + 0.5)  # ticks in the middle of the squares

    param_labels = [
        f"{name} [{unit}]" if unit else name
        for name, unit in zip(param_names, param_units, strict=True)
    ]
    ax.set_xticklabels(param_labels, size=font_size, rotation=label_rotation)
    ax.set_yticklabels(param_labels, size=font_size)
    ax.xaxis.set_ticks_position("bottom")
    ax.xaxis.set_label_position("bottom")

    ax.set_title(title, fontsize=1.4 * font_size)


def _format_number_as_latex(x, digits=2) -> str:
    """Format number in LaTeX style.

    Parameters
    ----------
    x : float
        Number to format.
    digits : int
        Number of significant digits.

    Returns
    -------
    str
        Formatted number.
    """

    if np.isnan(x):
        return "NaN"

    fmt = "{:." + str(digits) + "e}"
    a, n = fmt.format(x).split("e")

    if int(n) == 0:
        fmt = "{:." + str(digits) + "f}"
        return f"${fmt.format(x)}$"

    return f"${a}$ \n $\\times 10^{{{int(n)}}}$"



"""
Benchmark of inner products and other functions used by fim.
"""

import jax
import jax.numpy as jnp
import numpy as np

from wejax.fim import inner_product

N = 10000
M = 10
D = 2
a = np.random.randn(M, N, D)
b = np.random.randn(M, N, D)[..., [0], :, :]
DF = 0.1
noise_psd = np.ones((N, D, D))


def inner_product_nojax(a, b, df, noise_cov, should_invert_noise_cov, sum_func):
    if should_invert_noise_cov:
        inv_noise_cov = np.linalg.inv(noise_cov)  # (..., Nf, Nc, Nc)
    else:
        inv_noise_cov = noise_cov  # (..., Nf, Nc, Nc)
    a, b = np.broadcast_arrays(a, b)  # Contract over the channel axis
    subscripts = "...i, ...ij, ...j -> ..."
    contracted = np.einsum(subscripts, a, inv_noise_cov, np.conj(b))
    summed = 4 * np.real(
        sum_func(contracted, dx=df, axis=-1)
    )  # Sum over the frequencies
    return summed


def jax_sum(integrand, dx, axis):
    return jnp.sum(integrand, axis=axis) * dx


sum_ip = lambda: inner_product(
    a, b, df=DF, noise_cov=noise_psd, should_invert_noise_cov=False, sum_func=jax_sum
).block_until_ready()
trapz_ip = lambda: inner_product(
    a,
    b,
    df=DF,
    noise_cov=noise_psd,
    should_invert_noise_cov=False,
    sum_func=jnp.trapezoid,
).block_until_ready()
nojax_ip = lambda: inner_product_nojax(
    a,
    b,
    df=DF,
    noise_cov=noise_psd,
    should_invert_noise_cov=False,
    sum_func=np.trapezoid,
)

# make a first evaluation
sum_ip()
trapz_ip()

size_str = f"size {M},{D},{N}"
__benchmarks__ = [
    (
        sum_ip,
        trapz_ip,
        f"inner product with jax: using trapz instead of sum; {size_str}",
    ),
    (nojax_ip, trapz_ip, f"inner product with trapz: using jax; {size_str}"),
]



"""Test the jacobian module."""

import jax
import jax.numpy as jnp
import numpy as np
from jax import Array

from wejax.jacobian import autograd, findiff

jax.config.update("jax_enable_x64", True)


def _template_1(a: float) -> Array:
    """Univariate, single-channel, 10-frequency-point template."""
    return jnp.tile(a**2, (10, 1))


def _jacobian_1(a: float) -> Array:
    """Jacobian for :func:`_template_1`."""
    return jnp.tile(2 * a, (10, 1, 1))


def _template_2(a: float) -> Array:
    """Univariate, 3-channel, 10-frequency-point template."""
    channels = jnp.stack([2 * a**2 - 5 * a + 10, jnp.sin(a), jnp.exp(1j * a)])
    return jnp.tile(channels, (10, 1))


def _jacobian_2(a: float) -> Array:
    """Jacobian for :func:`_template_2`."""
    jac1 = 4 * a - 5
    jac2 = jnp.cos(a)
    jac3 = 1j * jnp.exp(1j * a)
    jac = jnp.stack([jac1, jac2, jac3])[:, None]
    return jnp.tile(jac, (10, 1, 1))


def _template_3(a: float, b: float) -> Array:
    """Bivariate, 2-channel, 10-frequency-point template."""
    channels = jnp.stack([a**2, 3 * a * b**2])
    return jnp.tile(channels, (10, 1))


def _jacobian_3(a: float, b: float) -> Array:
    """Jacobian for :func:`_template_3`."""
    jac1 = [2 * a, 0]
    jac2 = [3 * b**2, 6 * a * b]
    jac = jnp.array([jac1, jac2])
    return jnp.tile(jac, (10, 1, 1))


def test_autograd_jacobian() -> None:
    """Test the autograd_jacobian function."""

    jac_1 = autograd(_template_1, argnums=(0,))

    result = jac_1(42.0)
    expected = _jacobian_1(42.0)
    np.testing.assert_array_equal(result, expected)

    result = jax.vmap(jac_1)(jnp.array([1.0, 2.0]))
    expected = jnp.stack([_jacobian_1(1.0), _jacobian_1(2.0)])
    np.testing.assert_array_equal(result, expected)

    jac_2 = autograd(_template_2, argnums=(0,))

    result = jac_2(42.0)
    expected = _jacobian_2(42.0)
    np.testing.assert_array_equal(result, expected)

    result = jax.vmap(jac_2)(jnp.array([1.0, 2.0]))
    expected = jnp.stack([_jacobian_2(1.0), _jacobian_2(2.0)])
    np.testing.assert_array_equal(result, expected)

    jac_3 = autograd(_template_3, argnums=(0, 1))

    result = jac_3(42.0, 43.0)
    expected = _jacobian_3(42.0, 43.0)
    np.testing.assert_array_equal(result, expected)


def test_findiff_jacobian() -> None:
    """Test the findiff_jacobian function."""

    jac_1 = findiff(_template_1, argnums=(0,), steps=1e-6)

    result = jac_1(42.0)
    expected = _jacobian_1(42.0)
    np.testing.assert_array_almost_equal(result, expected, decimal=5)

    result = jax.vmap(jac_1)(jnp.array([1.0, 2.0]))
    expected = jnp.stack([_jacobian_1(1.0), _jacobian_1(2.0)])
    np.testing.assert_array_almost_equal(result, expected, decimal=5)

    jac_2 = findiff(_template_2, argnums=(0,), steps=1e-6)

    result = jac_2(42.0)
    expected = _jacobian_2(42.0)
    np.testing.assert_array_almost_equal(result, expected)

    result = jax.vmap(jac_2)(jnp.array([1.0, 2.0]))
    expected = jnp.stack([_jacobian_2(1.0), _jacobian_2(2.0)])
    np.testing.assert_array_almost_equal(result, expected, decimal=5)

    jac_3 = findiff(_template_3, argnums=(0, 1), steps=(1e-3, 1e-5))

    result = jac_3(42.0, 43.0)
    expected = _jacobian_3(42.0, 43.0)
    np.testing.assert_array_almost_equal(result, expected, decimal=5)



r"""
Templates and Jacobians
=======================

This module provides functions to evaluate the Jacobian of a template function.

Templates
---------

Template functions model the observed data. In this framework, we assume they
are given in the Fourier domain as complex-valued functions of the model
parameters and frequencies,

.. math::

    h_i(\theta)(f),

where :math:`i` is the channel (observable) index, :math:`\theta` is a set of
model parameters, and :math:`f` is the frequency vector.

The implementation of the template function is expected to be in JAX. It should
be a function of the model parametrers as positional arguments and return an
array of complex values for all frequencies and channels.

.. autoclass:: Template

Jacobians
---------

Given a template function :math:`h_i(\theta, f)` that depends on a set of
parameters :math:`\theta` and frequencies :math:`f`, the Jacobian is a
matrix of partial derivatives of the template with respect to the model
parameters,

.. math::

    J_{ij}(\theta)(f) = \frac{\partial h_i}{\partial \theta_j}(\theta)(f).

The implementation of the Jacobian function is expected to be in JAX. It should
be a function of the model parameters as positional arguments and return an
array of complex values for all frequencies, channels, and parameters (along
which the partial derivatives are computed).

.. autoclass:: Jacobian

This module provides convenience functions to compute Jacobians using automatic
differentiation and finite differences. It also provides a function to mix
multiple Jacobians along the parameter axis.

Automatic differentiation
^^^^^^^^^^^^^^^^^^^^^^^^^

The :func:`autograd` function computes the Jacobian using automatic
differentiation. This function uses JAX forward-mode automatic differentiation
to compute the Jacobian of the template function.

.. note:: Vectorization

    To vectorize the computation of the Jacobian for multiple sets of
    parameters, use :func:`jax.vmap` on the non-vectorized Jacobian function, to
    avoid computing partial derivatives for each parameter array element.

.. autofunction:: autograd

Finite differences
^^^^^^^^^^^^^^^^^^

The :func:`findiff` function computes the Jacobian using finite differences.
This function uses numerical differentiation to compute the Jacobian of the
template function.

The function requires a set of steps for numerical differentiation. These can
be either a scalar or an array with the same shape as the parameters to
differentiate. If a scalar, the same step is used for all parameters.

.. autofunction:: findiff

Mixing Jacobians
^^^^^^^^^^^^^^^^

The :func:`mix` function concatenates multiple Jacobians along the parameter
axis, based on the given argument indices.

.. autofunction:: mix

"""

from __future__ import annotations

from typing import Callable, Sequence, TypeAlias

import jax
import jax.numpy as jnp
import numpy as np
from jax import Array
from jax._src.typing import ArrayLike as JaxArrayLike
from numpy.typing import ArrayLike as NumpyArrayLike

Template: TypeAlias = Callable[..., Array]
r"""Template function.

Example
-------
Below is an example of a template function that models a single channel
with two parameters.

.. code-block:: python

    import jax.numpy as jnp
    from jax import Array

    freqs = jnp.linspace(0, 1, 10)

    def template(a: float, b: float) -> Array:
        channel = a**2 * freqs + 42j * jnp.cos(a) * b**2
        return jnp.stack([channel], axis=-1)

    template(1.0, 2.0)

This template can be vectorized for efficient computation of multiple sets
of parameters using :func:`jax.vmap` (see `JAX documentation
<https://docs.jax.dev/en/latest/_autosummary/jax.vmap.html>`_).

.. code-block:: python

    from jax import vmap

    a = jnp.array([1.0, 2.0, 3.0])
    b = jnp.array([4.0, 5.0, 6.0])
    vtemplate = vmap(template, in_axes=(0, 0), out_axes=0)
    vtemplate(a, b)

Parameters
----------
*args : Array-like
    Model parameters as positional arguments.

Returns
-------
Complex-valued array
    Template values.

    The returned shape is often ``(..., Nf, Nc)``, with ``Nf`` is the number
    of frequencies, ``Nc`` is the number of channels. The ellipsis
    represents any number of additional dimensions.
"""

Jacobian: TypeAlias = Callable[..., Array]
r"""Jacobian function.

    Example
    -------
    .. code-block:: python

        import jax.numpy as jnp
        from jax import Array

        freqs = jnp.linspace(0, 1, 10)

        def jac(a: float, b: float) -> Array:
            # Compute Jacobian elements
            jac_0 = jnp.stack([2 * a * freqs - 42j * b**2 * jnp.sin(a)], axis=-1)
            jac_1 = jnp.stack([84j * jnp.cos(a) * b * jnp.ones_like(freqs)], axis=-1)

            return jnp.stack([jac_0, jac_1], axis=-1)

        jac(1.0, 2.0)

    This Jacobian can be vectorized for efficient computation of multiple sets
    of parameters using :func:`jax.vmap` (see `JAX documentation
    <https://docs.jax.dev/en/latest/_autosummary/jax.vmap.html>`_).

    .. code-block:: python

        from jax import vmap

        a = jnp.array([1.0, 2.0, 3.0])
        b = jnp.array([4.0, 5.0, 6.0])
        vjac = vmap(jac, in_axes=(0, 0), out_axes=0)
        vjac(a, b)

    Parameters
    ----------
    *args : Array-like
        Model parameters as positional arguments.

    Returns
    -------
    Complex-valued array
        Template Jacobian values.

        The returned shape is often ``(..., Nf, Nc, Np)``, with ``Nf`` is the
        number of frequencies, ``Nc`` is the number of channels, and ``Np`` the
        number of parameters. The ellipsis represents any number of additional
        dimensions.
    """


def autograd(template: Template, argnums: Sequence[int]) -> Jacobian:
    r"""Compute the Jacobian of a function using automatic differentiation.

    This function use JAX forward-mode (column-wise :func:`jax.jacfwd`)
    automatic differentiation to compute the Jacobian of the template function.

    Example
    -------
    >>> from wejax.jacobian import autograd
    >>> jac = autograd(template, argnums=(0, 1))
    >>> jac(1.0, 2.0)
    Array([[[[...]]]], dtype=complex64)

    Parameters
    ----------
    template : :class:`Template`
        Template function, implemented in JAX.
    argnums : Sequence[int]
        Indices of the positional arguments to differentiate with respect to.

    Returns
    -------
    :class:`Jacobian`
        Jacobian function.
    """

    def jac(*args: JaxArrayLike) -> Array:
        return jnp.stack(
            jax.jacfwd(template, argnums=argnums)(*args),
            axis=-1,
        )

    return jac


def findiff(
    template: Template, argnums: Sequence[int], steps: NumpyArrayLike
) -> Jacobian:
    r"""Compute the Jacobian of a function using finite differences.

    This function use numerical differentiation to compute the Jacobian of the
    template function,

    .. math::

        J_{ij}(\theta)(f) = \frac{h_i(\theta + \delta_j)(f) - h_i(\theta -
        \delta_j)(f)}{2 \delta_j},

    where :math:`h_i(\theta)` is the template function, :math:`\theta` is the
    model parameters, :math:`f` is the frequency vector, and :math:`\delta_j`
    is the step for numerical differentiation.

    Example
    -------
    >>> from wejax.jacobian import findiff
    >>> jac = findiff(template, argnums=(0, 1), steps=(1e-6, 2e-6))
    >>> jac(1.0, 2.0)
    Array([[[[...]]]], dtype=complex64)

    Parameters
    ----------
    template : :class:`Template`
        Template function, implemented in JAX.
    argnums : Sequence[int]
        Indices of the positional arguments to differentiate with respect to.
    steps: Array-like
        Steps for numerical differentiation.

        These can be either a scalar or an array with the same shape as the
        parameters to differentiate. If a scalar, the same step is used for
        all parameters.

    Returns
    -------
    :class:`Jacobian`
        Jacobian function.
    """
    # Broadcast steps
    n_params = len(argnums)
    steps = np.broadcast_to(steps, (n_params,))

    def jac(*args: JaxArrayLike) -> Array:

        # Broadcast and stack parameters
        params = jnp.stack(
            jnp.broadcast_arrays(*args),
            axis=0,
        )

        # Compute finite differences per parameter
        findiffs = []
        for i, argnum in enumerate(argnums):
            left = params.at[argnum].subtract(steps[i])
            right = params.at[argnum].add(steps[i])
            diff = (template(*right) - template(*left)) / (2 * steps[i])
            findiffs.append(diff)

        return jnp.stack(findiffs, axis=-1)

    return jac


def mix(
    jacobians: Sequence[Jacobian],
    argnums: Sequence[Sequence[int]],
    sort_args: bool = False,
) -> Jacobian:
    r"""Concatenate Jacobians along the parameter axis.

    This function concatenates multiple Jacobians along the parameter axis,
    based on the given argument indices.

    Example
    -------
    >>> from wejax.jacobian import autograd, findiff, mix
    >>> jac1 = autograd(template1, argnums=(1,))
    >>> jac2 = findiff(template2, argnums=(0,), steps=1e-6)
    >>> jac = mix([jac1, jac2], argnums=[(1,), (0,)])
    >>> associated_argnums = (0, 2, 1,)

    You can sort the arguments before concatenating the Jacobians by setting
    ``sort_args=True``.

    >>> jac = mix([jac1, jac2], argnums=[(1,), (0,)], sort_args=True)
    >>> associated_argnums = (0, 1)

    Parameters
    ----------
    jacobians : Sequence[Jacobian]
        List of Jacobian functions.
    argnums : Sequence[Sequence[int]]
        List of lists of indices of argument indices attached to each Jacobian,
        in the same order as ``jacobians``.
    sort_args : bool, optional
        Whether to sort the arguments (i.e., Jacobian columns) before
        concatenating the Jacobians.

    Returns
    -------
    :class:`Jacobian`
        Mixed Jacobian function.
    """
    # Make sure indices are not repeated
    flat_argnums = [argnum for indices in argnums for argnum in indices]
    if len(flat_argnums) != len(set(flat_argnums)):
        raise ValueError("Indices in `argnums` must not be repeated.")

    # For each parameter, find out which Jacobian it belongs to and its index
    # That will allow us to concatenate the Jacobians along the parameter axis
    param_indices: list[tuple[int, int]]
    if sort_args:
        # To sort the arguments, prepare a list with placeholder tuples and
        # then populate it with the actual indices
        param_indices_temp: list[tuple[int, int] | None] = [None] * len(flat_argnums)
        for i, indices in enumerate(argnums):
            for j, index in enumerate(indices):
                param_indices_temp[index] = (i, j)
        # Assert that all indices were populated
        param_indices = [x for x in param_indices_temp if x is not None]
        if param_indices_temp != param_indices:
            raise RuntimeError("Could not sort the arguments.")
    else:
        # Here it's easier: just go through the argnums list and populate the
        # param_indices list with the actual Jacobian and parameter indices
        param_indices = [(i, j) for i, indices in enumerate(argnums) for j in indices]

    def jac(*args: JaxArrayLike) -> Array:
        jac_values = [jacobian(*args) for jacobian in jacobians]
        jac_columns = [jac_values[i][..., j] for i, j in param_indices]
        return jnp.stack(jac_columns, axis=-1)

    return jac



"""
Benchmark of jacobian related functions.
"""

import jax
import jax.numpy as jnp
import numpy as np

from wejax.jacobian import autograd, findiff

jax.config.update("jax_enable_x64", True)


class Single:
    """Single evaluation of a polynomial."""

    def _set_dim(self):
        n_freqs = 100_000
        df = 0.01
        self.freqs = np.arange(1, n_freqs + 1) * df

    def template(self, a, b):
        """A simple polynomial template"""
        channel = a**2 * self.freqs + 42j * jnp.cos(a) * b**2
        return jnp.stack([channel], axis=-1)

    def template_nojax(self, a, b):
        """A simple polynomial template, without jax"""
        channel = a**2 * self.freqs + 42j * np.cos(a) * b**2
        return np.stack([channel], axis=-1)

    def __init__(self):

        self._set_dim()
        a = np.ones((1,))
        b = np.ones((1,))

        self.apply_autograd = lambda: autograd(self.template, argnums=(0, 1))(a, b)
        self.apply_findiff = lambda: findiff(
            self.template, argnums=(0, 1), steps=[0.01, 0.1]
        )(a, b)
        self.apply_findiff_nojax = lambda: findiff(
            self.template_nojax, argnums=(0, 1), steps=[0.01, 0.1]
        )(a, b)

        # make a first evaluation
        self.apply_findiff()
        self.apply_autograd()

    def get_default_benchmark(self):
        """Return richbench inputs"""
        return [
            (
                self.apply_findiff,
                self.apply_autograd,
                "jac: using autograd instead of finite differences;",
            ),
            (
                self.apply_findiff_nojax,
                self.apply_findiff,
                "jac fin. diff.: using jax;",
            ),
        ]


class Vectorized(Single):
    """Vectorized evaluation of a polynomial using vmap or list comprehension."""

    def __init__(self):
        # pylint: disable=super-init-not-called
        Single._set_dim(self)
        n_col = 50
        va = jnp.ones((n_col,))
        vb = jnp.ones((n_col,))

        jacobian_findiff = findiff(self.template, argnums=(0, 1), steps=[0.01, 0.1])
        self.vect_findiff = lambda: jax.vmap(jacobian_findiff, in_axes=(0, 0))(
            va, vb
        ).block_until_ready()
        self.clist_findiff = lambda: [
            jacobian_findiff(_a, _b) for _a, _b in zip(va, vb)
        ]

        jacobian_autograd = autograd(self.template, argnums=(0, 1))
        self.vect_autograd = lambda: jax.vmap(jacobian_autograd, in_axes=(0, 0))(
            va, vb
        ).block_until_ready()
        self.clist_autograd = lambda: [
            jacobian_autograd(_a, _b) for _a, _b in zip(va, vb)
        ]
        # pylint: disable=not-callable
        self.jit_autograd = lambda: jax.jit(
            jax.vmap(jacobian_autograd, in_axes=(0, 0))
        )(va, vb).block_until_ready()

    def get_default_benchmark(self):
        """Return richbench inputs"""
        return [
            (
                self.clist_findiff,
                self.vect_findiff,
                "jac fin. diff.: vmap instead of list comp.;",
            ),
            (
                self.clist_autograd,
                self.vect_autograd,
                "jac autograd: vmap instead of list comp.;",
            ),
            (
                self.vect_autograd,
                self.jit_autograd,
                "jac autograd: jit on top of vmap;",
            ),
        ]


class SingleFFT:
    """A template with an fft"""

    def _set_dim(self):
        n_t = 100_000
        dt = 5
        self.times = np.arange(1, n_t + 1, dt)

    def wfm_template(self, ampl, phase):
        """A waveform template, with FFT"""
        phase = phase * self.times
        plus = -jnp.cos(phase) * ampl
        cross = -jnp.sin(phase) * 2.0 * ampl
        return jnp.stack([jnp.fft.fft(plus), jnp.fft.fft(cross)], axis=1)

    def wfm_template_nojax(self, ampl, phase):
        """A waveform template, with FFT, without JAX"""
        phase = phase * self.times
        plus = -np.cos(phase) * ampl
        cross = -np.sin(phase) * 2.0 * ampl
        return np.stack([np.fft.fft(plus), np.fft.fft(cross)], axis=1)

    def __init__(self):
        self._set_dim()

        ampl = np.ones((1,))
        phase = np.ones((1,)) * np.pi / 6.0

        self.fft_autograd = lambda: autograd(self.wfm_template, argnums=(0, 1))(
            ampl, phase
        )
        self.fft_findiff = lambda: findiff(
            self.wfm_template, argnums=(0, 1), steps=[0.1, 0.01]
        )(ampl, phase)
        self.fft_findiff_nojax = lambda: findiff(
            self.wfm_template_nojax, argnums=(0, 1), steps=[0.1, 0.01]
        )(ampl, phase)

        # make a first evaluation
        self.fft_findiff()
        self.fft_autograd()

    def get_default_benchmark(self):
        """Return richbench inputs"""
        return [
            (
                self.fft_findiff,
                self.fft_autograd,
                "jac fft: using autograd instead of finite differences; ",
            ),
            (
                self.fft_findiff_nojax,
                self.fft_findiff,
                "jac fft fin. diff.: using jax; ",
            ),
        ]


class VectorizedFFT(SingleFFT):
    """
    Vectorized usage of a template with an fft
    """

    def __init__(self):
        # pylint: disable=super-init-not-called
        self._set_dim()

        n_col = 50
        vampl = jnp.ones((n_col,))
        vphase = jnp.ones((n_col,)) * np.pi / 6.0

        fft_jacobian_findiff = findiff(
            self.wfm_template, argnums=(0, 1), steps=[0.1, 0.01]
        )
        self.fft_vect_findiff = lambda: jax.vmap(fft_jacobian_findiff, in_axes=(0, 0))(
            vampl, vphase
        ).block_until_ready()
        self.fft_clist_findiff = lambda: [
            fft_jacobian_findiff(_a, _b) for _a, _b in zip(vampl, vphase)
        ]

        fft_jacobian_autograd = autograd(self.wfm_template, argnums=(0, 1))
        self.fft_vect_autograd = lambda: jax.vmap(
            fft_jacobian_autograd, in_axes=(0, 0)
        )(vampl, vphase).block_until_ready()
        self.fft_clist_autograd = lambda: [
            fft_jacobian_autograd(_a, _b) for _a, _b in zip(vampl, vphase)
        ]
        # pylint: disable=not-callable
        self.fft_jit_autograd = lambda: jax.jit(
            jax.vmap(fft_jacobian_autograd, in_axes=(0, 0))
        )(vampl, vphase).block_until_ready()

    def get_default_benchmark(self):
        """Return richbench inputs"""
        return [
            (
                self.fft_clist_findiff,
                self.fft_vect_findiff,
                "jac fft fin. diff.: vmap instead of list comp.;",
            ),
            (
                self.fft_clist_autograd,
                self.fft_vect_autograd,
                "jac fft autograd: vmap instead of list comp.;",
            ),
            (
                self.fft_vect_autograd,
                self.fft_jit_autograd,
                "jac fft autograd: jit on top of vmap;",
            ),
        ]


__benchmarks__ = (
    Single().get_default_benchmark()
    + Vectorized().get_default_benchmark()
    + SingleFFT().get_default_benchmark()
    + VectorizedFFT().get_default_benchmark()
)
